import asyncio
from asyncio import CancelledError
from typing import Any, Awaitable, Callable, Dict, TypeVar, Union, Optional

from cilroy.models import SerializableState
from cilroy.retry.base import Retrier
from kilroy_server_py_utils import Configurable, Parameter, classproperty

T = TypeVar("T")


class State(SerializableState):
    tries: Optional[int] = 3
    timeout: float = 1.0


class SimpleRetrier(Retrier, Configurable[State]):
    class TriesParameter(Parameter[State, Optional[int]]):
        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": ["integer", "null"],
                "minimum": 1,
                "title": cls.pretty_name,
                "default": 3,
            }

    class TimeoutParameter(Parameter[State, float]):
        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "number",
                "minimum": 0.0,
                "title": cls.pretty_name,
                "default": 1.0,
            }

    async def retry(
        self,
        f: Union[Callable[..., T], Callable[..., Awaitable[T]]],
        *args,
        _on_retry: Optional[Callable[[int, float], None]] = None,
        **kwargs,
    ) -> T:
        is_coroutine = asyncio.iscoroutinefunction(f)

        async def call(f_, *args_, **kwargs_) -> T:
            if is_coroutine:
                return await f_(*args_, **kwargs_)
            else:
                return f_(*args_, **kwargs_)

        i = 0
        exceptions = []

        while True:
            async with self.state.read_lock() as state:
                tries = state.tries
                timeout = state.timeout

            try:
                return await call(f, *args, **kwargs)
            except CancelledError:
                raise
            except Exception as e:
                exceptions.append(e)
                i += 1

                if tries is not None and i >= tries:
                    raise Exception(*exceptions)

                if _on_retry is not None:
                    _on_retry(i, timeout)

                await asyncio.sleep(timeout)



import logging
from enum import Enum

import typer


class Verbosity(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class TyperLoggerHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        fg = None
        bg = None
        if record.levelno == logging.DEBUG:
            fg = typer.colors.BLACK
        elif record.levelno == logging.INFO:
            fg = typer.colors.BLUE
        elif record.levelno == logging.WARNING:
            fg = typer.colors.YELLOW
        elif record.levelno == logging.CRITICAL:
            fg = typer.colors.RED
        elif record.levelno == logging.ERROR:
            fg = typer.colors.RED
        typer.secho(self.format(record), bg=bg, fg=fg)


def configure(verbosity) -> None:
    logging.basicConfig(
        format="%(levelname)s\t%(message)s",
        level=verbosity.value,
        handlers=[TyperLoggerHandler()],
    )



from datetime import datetime, time, timedelta, timezone
from math import ceil
from typing import Iterable

from croniter import croniter


def next_time(base: datetime, interval: timedelta) -> datetime:
    utcbase = base.astimezone(timezone.utc)
    utcnow = datetime.now(timezone.utc)
    return utcbase + interval * ceil((utcnow - utcbase) / interval)


def next_time_from_hours(hours: Iterable[time]) -> datetime:
    now = datetime.now(timezone.utc)
    return min(
        next_time(
            datetime.combine(now.date(), hour, tzinfo=timezone.utc),
            timedelta(days=1),
        )
        for hour in hours
    )


def next_time_from_rule(rule: str) -> datetime:
    now = datetime.now(timezone.utc)
    return croniter(rule, now).get_next(datetime)


def seconds_until(dt: datetime) -> float:
    return (dt - datetime.now(timezone.utc)).total_seconds()


def utcmidnight() -> datetime:
    now = datetime.now(timezone.utc)
    return datetime.combine(now.date(), time(), tzinfo=timezone.utc)



import asyncio
from datetime import time
from pathlib import Path
from typing import Any, AsyncIterable, Awaitable, Callable, Dict, List

from cilroy.models import SerializableState
from cilroy.schedulers.base import Scheduler
from cilroy.utils import seconds_until, next_time_from_hours
from kilroy_server_py_utils import Configurable, Parameter, classproperty


class State(SerializableState):
    hours: List[time] = []


class TimetableScheduler(Scheduler, Configurable[State]):
    class HoursParameter(Parameter[State, List[str]]):
        @classmethod
        async def _get(cls, state: State) -> List[str]:
            return [hour.isoformat(timespec="seconds") for hour in state.hours]

        @classmethod
        async def _set(
            cls, state: State, value: List[str]
        ) -> Callable[[], Awaitable]:
            original_value = state.hours

            async def undo():
                state.hours = original_value

            state.hours = sorted([time.fromisoformat(hour) for hour in value])
            return undo

        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "array",
                "items": {"type": "string", "format": "time"},
                "title": cls.pretty_name,
                "default": [],
            }

    @classmethod
    async def _save_state(cls, state: State, directory: Path) -> None:
        with open(directory / "state.json", "w") as f:
            f.write(state.json())

    async def _load_saved_state(self, directory: Path) -> State:
        with open(directory / "state.json", "r") as f:
            return State.parse_raw(f.read())

    async def wait(self) -> AsyncIterable[None]:
        while True:
            async with self.state.read_lock() as state:
                hours = state.hours

            score_time = next_time_from_hours(hours)
            await asyncio.sleep(seconds_until(score_time))
            yield



from datetime import datetime
from typing import Any, Awaitable, Callable, Dict, Optional, Type

from cilroy.controller.state import State
from cilroy.retry import Retrier, ExponentialBackoffRetrier
from cilroy.schedulers import Scheduler, IntervalScheduler
from kilroy_server_py_utils import (
    CategorizableBasedParameter,
    OptionalParameter,
    Parameter,
    classproperty,
)


class ScrapBeforeParameter(OptionalParameter[State, str]):
    @classmethod
    async def _get(cls, state: State) -> Optional[str]:
        if state.offline.scrap_before is None:
            return None
        return state.offline.scrap_before.isoformat()

    @classmethod
    async def _set(
        cls, state: State, value: Optional[str]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.scrap_before

        async def undo():
            state.offline.scrap_before = original_value

        state.offline.scrap_before = (
            datetime.fromisoformat(value) if value is not None else None
        )
        return undo

    # noinspection PyMethodParameters
    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["string", "null"],
            "format": "date-time",
            "title": cls.pretty_name,
            "default": None,
        }


class ScrapAfterParameter(OptionalParameter[State, str]):
    @classmethod
    async def _get(cls, state: State) -> Optional[str]:
        if state.offline.scrap_after is None:
            return None
        return state.offline.scrap_after.isoformat()

    @classmethod
    async def _set(
        cls, state: State, value: Optional[str]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.scrap_after

        async def undo():
            state.offline.scrap_after = original_value

        state.offline.scrap_after = (
            datetime.fromisoformat(value) if value is not None else None
        )
        return undo

    # noinspection PyMethodParameters
    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["string", "null"],
            "format": "date-time",
            "title": cls.pretty_name,
            "default": None,
        }


class ScrapLimitParameter(OptionalParameter[State, int]):
    @classmethod
    async def _get(cls, state: State) -> Optional[int]:
        if state.offline.scrap_limit is None:
            return None
        return state.offline.scrap_limit

    @classmethod
    async def _set(
        cls, state: State, value: Optional[int]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.scrap_limit

        async def undo():
            state.offline.scrap_limit = original_value

        state.offline.scrap_limit = value if value is not None else None
        return undo

    # noinspection PyMethodParameters
    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["integer", "null"],
            "minimum": 0,
            "title": cls.pretty_name,
            "default": None,
        }


class PostSchedulerParameter(CategorizableBasedParameter[State, Scheduler]):
    @classmethod
    async def _get_categorizable(cls, state: State) -> Scheduler:
        return state.online.post_scheduler

    @classmethod
    async def _set_categorizable(cls, state: State, value: Scheduler) -> None:
        state.online.post_scheduler = value

    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return state.online.post_schedulers_params.get(category, {})

    # noinspection PyMethodParameters
    @classproperty
    def default_categorizable(cls) -> Type[Scheduler]:
        return IntervalScheduler

    # noinspection PyMethodParameters
    @classproperty
    def default_config(cls) -> Optional[Dict[str, Any]]:
        return {"interval": 3600}


class ScoreSchedulerParameter(CategorizableBasedParameter[State, Scheduler]):
    @classmethod
    async def _get_categorizable(cls, state: State) -> Scheduler:
        return state.online.score_scheduler

    @classmethod
    async def _set_categorizable(cls, state: State, value: Scheduler) -> None:
        state.online.score_scheduler = value

    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return state.online.score_schedulers_params.get(category, {})

    # noinspection PyMethodParameters
    @classproperty
    def default_categorizable(cls) -> Type[Scheduler]:
        return IntervalScheduler

    # noinspection PyMethodParameters
    @classproperty
    def default_config(cls) -> Optional[Dict[str, Any]]:
        return {"interval": 86400}


class AutosaveSchedulerParameter(
    CategorizableBasedParameter[State, Scheduler]
):
    @classmethod
    async def _get_categorizable(cls, state: State) -> Scheduler:
        return state.autosave.scheduler

    @classmethod
    async def _set_categorizable(cls, state: State, value: Scheduler) -> None:
        state.autosave.scheduler = value

    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return state.autosave.schedulers_params.get(category, {})

    # noinspection PyMethodParameters
    @classproperty
    def default_categorizable(cls) -> Type[Scheduler]:
        return IntervalScheduler

    # noinspection PyMethodParameters
    @classproperty
    def default_config(cls) -> Optional[Dict[str, Any]]:
        return {"interval": 3600}


class FeedLengthParameter(Parameter[State, int]):
    @classmethod
    async def _get(cls, state: State) -> int:
        return state.feed.length

    @classmethod
    async def _set(cls, state: State, value: int) -> Callable[[], Awaitable]:
        original_value = state.feed.length
        original_feed = state.feed.feed

        async def undo():
            state.feed.length = original_value
            state.feed.feed = original_feed

        state.feed.length = value
        state.feed.feed = state.feed.feed[-value:]
        return undo

    # noinspection PyMethodParameters
    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": "integer",
            "minimum": 1,
            "title": cls.pretty_name,
            "default": 100,
        }


class RetrierParameter(CategorizableBasedParameter[State, Retrier]):
    @classmethod
    async def _get_categorizable(cls, state: State) -> Retrier:
        return state.retry.retrier

    @classmethod
    async def _set_categorizable(cls, state: State, value: Retrier) -> None:
        state.retry.retrier = value

    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return state.retry.retriers_params.get(category, {})

    # noinspection PyMethodParameters
    @classproperty
    def default_categorizable(cls) -> Type[Retrier]:
        return ExponentialBackoffRetrier



from datetime import datetime
from typing import Any, Dict, Optional

from cilroy.models import SerializableModel
from cilroy.retry import ExponentialBackoffRetrier
from cilroy.schedulers import IntervalScheduler


class FaceParams(SerializableModel):
    host: str = "localhost"
    port: int = 10000


class ModuleParams(SerializableModel):
    host: str = "localhost"
    port: int = 11000


class OfflineParams(SerializableModel):
    scrap_before: Optional[datetime] = None
    scrap_after: Optional[datetime] = None
    scrap_limit: Optional[int] = None


class OnlineParams(SerializableModel):
    post_scheduler_type: str = IntervalScheduler.category
    post_schedulers_params: Dict[str, Dict[str, Any]] = {
        "interval": {"interval": 3600},
    }
    score_scheduler_type: str = IntervalScheduler.category
    score_schedulers_params: Dict[str, Dict[str, Any]] = {
        "interval": {"interval": 86400},
    }


class AutosaveParams(SerializableModel):
    scheduler_type: str = IntervalScheduler.category
    schedulers_params: Dict[str, Dict[str, Any]] = {
        "interval": {"interval": 3600},
    }


class FeedParams(SerializableModel):
    length: int = 100


class RetryParams(SerializableModel):
    retrier_type: str = ExponentialBackoffRetrier.category
    retriers_params: Dict[str, Dict[str, Any]] = {}


class Params(SerializableModel):
    face: FaceParams = FaceParams()
    module: ModuleParams = ModuleParams()
    offline: OfflineParams = OfflineParams()
    online: OnlineParams = OnlineParams()
    autosave: AutosaveParams = AutosaveParams()
    feed: FeedParams = FeedParams()
    retry: RetryParams = RetryParams()



from abc import ABC, abstractmethod
from typing import TypeVar, Callable, Awaitable, Union, Optional

from kilroy_server_py_utils import Categorizable, classproperty, normalize

T = TypeVar("T")


class Retrier(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("Retrier"))

    @abstractmethod
    async def retry(
        self,
        f: Union[Callable[..., T], Callable[..., Awaitable[T]]],
        *args,
        _on_retry: Optional[Callable[[int, float], None]] = None,
        **kwargs,
    ) -> T:
        pass



import asyncio
from pathlib import Path
from typing import Any, AsyncIterable, Awaitable, Callable, Dict

from cilroy.models import SerializableState
from cilroy.schedulers.base import Scheduler
from cilroy.utils import (
    seconds_until,
    next_time_from_rule,
)
from kilroy_server_py_utils import Configurable, Parameter, classproperty


class State(SerializableState):
    rule: str = "0 * * * *"


class CronScheduler(Scheduler, Configurable[State]):
    class RuleParameter(Parameter[State, str]):
        @classmethod
        async def _get(cls, state: State) -> str:
            return state.rule

        @classmethod
        async def _set(
            cls, state: State, value: str
        ) -> Callable[[], Awaitable]:
            original_value = state.rule

            async def undo():
                state.rule = original_value

            state.rule = value
            return undo

        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "string",
                "title": cls.pretty_name,
                "default": "0 * * * *",
            }

    @classmethod
    async def _save_state(cls, state: State, directory: Path) -> None:
        with open(directory / "state.json", "w") as f:
            f.write(state.json())

    async def _load_saved_state(self, directory: Path) -> State:
        with open(directory / "state.json", "r") as f:
            return State.parse_raw(f.read())

    async def wait(self) -> AsyncIterable[None]:
        while True:
            async with self.state.read_lock() as state:
                rule = state.rule

            score_time = next_time_from_rule(rule)
            await asyncio.sleep(seconds_until(score_time))
            yield



from cilroy.schedulers.base import Scheduler
from cilroy.schedulers.interval import IntervalScheduler
from cilroy.schedulers.timetable import TimetableScheduler
from cilroy.schedulers.cron import CronScheduler



from abc import ABC, abstractmethod
from typing import AsyncIterable

from kilroy_server_py_utils import Categorizable, classproperty, normalize


class Scheduler(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("Scheduler"))

    @abstractmethod
    def wait(self) -> AsyncIterable[None]:
        pass
